@@ -2,7 +2,8 @@
 
 module GHC.RTS.EventTypes where
 
-import Data.Word (Word16, Word32, Word64)
+import Data.Word (Word8, Word16, Word32, Word64)
+import Data.Binary
 
 -- EventType.
 type EventTypeNum = Word16
@@ -19,6 +20,15 @@ type BlockSize = Word32
 type RawThreadStopStatus = Word16
 type StringId = Word32
 type Capset   = Word32
+type PerfEventTypeNum = Word32
+type TaskId = Word64
+type PID = Word32
+
+newtype KernelThreadId = KernelThreadId { kernelThreadId :: Word64 }
+  deriving (Eq, Ord, Show)
+instance Binary KernelThreadId where
+  put (KernelThreadId tid) = put tid
+  get = fmap KernelThreadId get
 
 -- These types are used by Mercury events.
 type ParConjDynId = Word64
@@ -48,10 +58,16 @@ sz_block_event = fromIntegral (sz_event_type_num + sz_time + sz_block_size
     + sz_time + sz_cap)
 sz_pid :: EventTypeSize
 sz_pid = 4
+sz_taskid :: EventTypeSize
+sz_taskid = 8
+sz_kernel_tid :: EventTypeSize
+sz_kernel_tid = 8
 sz_th_stop_status :: EventTypeSize
 sz_th_stop_status = 2
 sz_string_id :: EventTypeSize
 sz_string_id = 4
+sz_perf_num :: EventTypeSize
+sz_perf_num = 4
 
 -- Sizes for Mercury event fields.
 sz_par_conj_dyn_id :: EventTypeSize
@@ -148,6 +164,17 @@ data EventInfo
   | SparkFizzle        { }
   | SparkGC            { }
 
+  -- tasks
+  | TaskCreate         { taskId :: TaskId,
+                         cap :: {-# UNPACK #-}!Int,
+                         tid :: {-# UNPACK #-}!KernelThreadId
+                       }
+  | TaskMigrate        { taskId :: TaskId,
+                         cap :: {-# UNPACK #-}!Int,
+                         new_cap :: {-# UNPACK #-}!Int
+                       }
+  | TaskDelete         { taskId :: TaskId }
+
   -- garbage collection
   | RequestSeqGC       { }
   | RequestParGC       { }
@@ -218,10 +245,10 @@ data EventInfo
                        , env    :: [String]
                        }
   | OsProcessPid       { capset :: {-# UNPACK #-}!Capset
-                       , pid    :: {-# UNPACK #-}!Word32
+                       , pid    :: {-# UNPACK #-}!PID
                        }
   | OsProcessParentPid { capset :: {-# UNPACK #-}!Capset
-                       , ppid   :: {-# UNPACK #-}!Word32
+                       , ppid   :: {-# UNPACK #-}!PID
                        }
   | WallClockTime      { capset :: {-# UNPACK #-}!Capset
                        , sec    :: {-# UNPACK #-}!Word64
@@ -231,6 +258,7 @@ data EventInfo
   -- messages
   | Message            { msg :: String }
   | UserMessage        { msg :: String }
+  | UserMarker         { markername :: String }
 
   -- These events have been added for Mercury's benifit but are generally
   -- useful.
@@ -273,6 +301,18 @@ data EventInfo
   | MerCapSleeping
   | MerCallingMain
 
+  -- perf events
+  | PerfName           { perfNum :: {-# UNPACK #-}!PerfEventTypeNum
+                       , name    :: String
+                       }
+  | PerfCounter        { perfNum :: {-# UNPACK #-}!PerfEventTypeNum
+                       , tid     :: {-# UNPACK #-}!KernelThreadId
+                       , period  :: {-# UNPACK #-}!Word64
+                       }
+  | PerfTracepoint     { perfNum :: {-# UNPACK #-}!PerfEventTypeNum
+                       , tid     :: {-# UNPACK #-}!KernelThreadId
+                       }
+
   deriving Show
 
 --sync with ghc/includes/Constants.h
