@@ -1,8 +1,6 @@
 {-# LANGUAGE CPP,BangPatterns,PatternGuards #-}
 {-# OPTIONS_GHC -funbox-strict-fields -fwarn-incomplete-patterns #-}
 {-
- - Authors: Donnie Jones, Simon Marlow, Paul Bone, Duncan Coutts
- - Events.hs
  -   Parser functions for GHC RTS EventLog framework.
  -}
 
@@ -18,6 +16,8 @@ module GHC.RTS.Events (
        CapsetType(..),
        Timestamp,
        ThreadId,
+       TaskId,
+       KernelThreadId(..),
 
        -- * Reading and writing event logs
        readEventLogFromFile,
@@ -29,7 +29,11 @@ module GHC.RTS.Events (
 
        -- * Printing
        showEventInfo, showThreadStopStatus,
-       ppEventLog, ppEventType, ppEvent
+       ppEventLog, ppEventType, ppEvent,
+
+       -- * Perf events
+       nEVENT_PERF_NAME, nEVENT_PERF_COUNTER, nEVENT_PERF_TRACEPOINT,
+       sz_perf_num, sz_kernel_tid
   ) where
 
 {- Libraries. -}
@@ -263,7 +267,6 @@ standardParsers = [
       return WallClockTime{capset=cs,sec=s,nsec=ns}
   )),
 
-
  (VariableSizeParser EVENT_LOG_MSG (do -- (msg)
       num <- getE :: GetEvents Word16
       string <- getString num
@@ -274,6 +277,11 @@ standardParsers = [
       string <- getString num
       return UserMessage{ msg = string }
    )),
+    (VariableSizeParser EVENT_USER_MARKER (do -- (markername)
+      num <- getE :: GetEvents Word16
+      string <- getString num
+      return UserMarker{ markername = string }
+   )),
  (VariableSizeParser EVENT_PROGRAM_ARGS (do -- (capset, [arg])
       num <- getE :: GetEvents Word16
       cs <- getE
@@ -405,6 +413,25 @@ ghc7Parsers = [
  (simpleEvent EVENT_SPARK_FIZZLE   SparkFizzle),
  (simpleEvent EVENT_SPARK_GC       SparkGC),
 
+ (FixedSizeParser EVENT_TASK_CREATE (sz_taskid + sz_cap + sz_kernel_tid) (do  -- (taskID, cap, tid)
+      taskId <- getE :: GetEvents TaskId
+      cap    <- getE :: GetEvents CapNo
+      tid    <- getE :: GetEvents KernelThreadId
+      return TaskCreate{ taskId, cap = fromIntegral cap, tid }
+   )),
+ (FixedSizeParser EVENT_TASK_MIGRATE (sz_taskid + sz_cap*2) (do  -- (taskID, cap, new_cap)
+      taskId  <- getE :: GetEvents TaskId
+      cap     <- getE :: GetEvents CapNo
+      new_cap <- getE :: GetEvents CapNo
+      return TaskMigrate{ taskId, cap = fromIntegral cap
+                                , new_cap = fromIntegral new_cap
+                        }
+   )),
+ (FixedSizeParser EVENT_TASK_DELETE (sz_taskid) (do  -- (taskID)
+      taskId <- getE :: GetEvents TaskId
+      return TaskDelete{ taskId }
+   )),
+
  (FixedSizeParser EVENT_THREAD_WAKEUP (sz_tid + sz_cap) (do  -- (thread, other_cap)
       t <- getE
       oc <- getE :: GetEvents CapNo
@@ -544,6 +571,28 @@ mercuryParsers = [
 
  ]
 
+perfParsers = [
+ (VariableSizeParser EVENT_PERF_NAME (do -- (perf_num, name)
+      num     <- getE :: GetEvents Word16
+      perfNum <- getE
+      name    <- getString (num - sz_perf_num)
+      return PerfName{perfNum, name}
+   )),
+
+ (FixedSizeParser EVENT_PERF_COUNTER (sz_perf_num + sz_kernel_tid + 8) (do -- (perf_num, tid, period)
+      perfNum <- getE
+      tid     <- getE
+      period  <- getE
+      return PerfCounter{perfNum, tid, period}
+  )),
+
+ (FixedSizeParser EVENT_PERF_TRACEPOINT (sz_perf_num + sz_kernel_tid) (do -- (perf_num, tid)
+      perfNum <- getE
+      tid     <- getE
+      return PerfTracepoint{perfNum, tid}
+  ))
+ ]
+
 getData :: GetEvents Data
 getData = do
    db <- getE :: GetEvents Marker
@@ -586,8 +635,8 @@ getEventLog = do
         -}
         event_parsers = if is_ghc_6
                             then standardParsers ++ ghc6Parsers
-                            else standardParsers ++ ghc7Parsers ++
-                                mercuryParsers
+                            else standardParsers ++ ghc7Parsers
+                                 ++ mercuryParsers ++ perfParsers
         parsers = mkEventTypeParsers imap event_parsers
     dat <- runReaderT getData (EventParsers parsers)
     return (EventLog header dat)
@@ -703,6 +752,14 @@ showEventInfo spec =
           printf "spark fizzled"
         SparkGC ->
           printf "spark GCed"
+        TaskCreate taskId cap tid ->
+          printf "task 0x%x created on cap %d with OS kernel thread %d"
+                 taskId cap (kernelThreadId tid)
+        TaskMigrate taskId cap new_cap ->
+          printf "task 0x%x migrated from cap %d to cap %d"
+                 taskId cap new_cap
+        TaskDelete taskId ->
+          printf "task 0x%x deleted" taskId
         Shutdown ->
           printf "shutting down"
         WakeupThread thread otherCap ->
@@ -747,6 +804,8 @@ showEventInfo spec =
           msg
         UserMessage msg ->
           msg
+        UserMarker markername ->
+          printf "marker: %s" markername
         CapsetCreate cs ct ->
           printf "created capset %d of type %s" cs (show ct)
         CapsetDelete cs ->
@@ -799,6 +858,14 @@ showEventInfo spec =
           "Capability going to sleep"
         MerCallingMain ->
           "About to call the program entry point"
+        PerfName{perfNum, name} ->
+          printf "perf event %d named \"%s\"" perfNum name
+        PerfCounter{perfNum, tid, period} ->
+          printf "perf event counter %d incremented by %d in OS thread %d"
+                 perfNum (period + 1) (kernelThreadId tid)
+        PerfTracepoint{perfNum, tid} ->
+          printf "perf event tracepoint %d reached in OS thread %d"
+                 perfNum (kernelThreadId tid)
 
 showThreadStopStatus :: ThreadStopStatus -> String
 showThreadStopStatus HeapOverflow   = "heap overflow"
@@ -926,10 +993,14 @@ eventTypeNum e = case e of
     SparkSteal    {} -> EVENT_SPARK_STEAL
     SparkFizzle   {} -> EVENT_SPARK_FIZZLE
     SparkGC       {} -> EVENT_SPARK_GC
+    TaskCreate  {} -> EVENT_TASK_CREATE
+    TaskMigrate {} -> EVENT_TASK_MIGRATE
+    TaskDelete  {} -> EVENT_TASK_DELETE
     Message {} -> EVENT_LOG_MSG
     Startup {} -> EVENT_STARTUP
     EventBlock {} -> EVENT_BLOCK_MARKER
     UserMessage {} -> EVENT_USER_MSG
+    UserMarker  {} -> EVENT_USER_MARKER
     GCIdle {} -> EVENT_GC_IDLE
     GCWork {} -> EVENT_GC_WORK
     GCDone {} -> EVENT_GC_DONE
@@ -968,6 +1039,14 @@ eventTypeNum e = case e of
     MerReleaseThread _ -> EVENT_MER_RELEASE_CONTEXT
     MerCapSleeping -> EVENT_MER_ENGINE_SLEEPING
     MerCallingMain -> EVENT_MER_CALLING_MAIN
+    PerfName       {} -> nEVENT_PERF_NAME
+    PerfCounter    {} -> nEVENT_PERF_COUNTER
+    PerfTracepoint {} -> nEVENT_PERF_TRACEPOINT
+
+nEVENT_PERF_NAME, nEVENT_PERF_COUNTER, nEVENT_PERF_TRACEPOINT :: EventTypeNum
+nEVENT_PERF_NAME = EVENT_PERF_NAME
+nEVENT_PERF_COUNTER = EVENT_PERF_COUNTER
+nEVENT_PERF_TRACEPOINT = EVENT_PERF_TRACEPOINT
 
 putEvent :: Event -> PutEvents ()
 putEvent (Event t spec) = do
@@ -1097,13 +1176,26 @@ putEventSpec EndGC = do
 putEventSpec GlobalSyncGC = do
     return ()
 
+putEventSpec (TaskCreate taskId cap tid) = do
+    putE taskId
+    putCap cap
+    putE tid
+
+putEventSpec (TaskMigrate taskId cap new_cap) = do
+    putE taskId
+    putCap cap
+    putCap new_cap
+
+putEventSpec (TaskDelete taskId) = do
+    putE taskId
+
 putEventSpec GCStatsGHC{..} = do
     putE heapCapset
-    putE gen
+    putE (fromIntegral gen :: Word16)
     putE copied
     putE slop
     putE frag
-    putE parNThreads
+    putE (fromIntegral parNThreads :: Word32)
     putE parMaxCopied
     putE parTotCopied
 
@@ -1121,7 +1213,7 @@ putEventSpec HeapLive{..} = do
 
 putEventSpec HeapInfoGHC{..} = do
     putE heapCapset
-    putE gens
+    putE (fromIntegral gens :: Word16)
     putE maxHeapSize
     putE allocAreaSize
     putE mblockSize
@@ -1159,19 +1251,19 @@ putEventSpec (CapsetRemoveCap cs cp) = do
     putCap cp
 
 putEventSpec (RtsIdentifier cs rts) = do
-    putE (fromIntegral (length rts) + 4 :: Word16)
+    putE (fromIntegral (length rts) + sz_capset :: Word16)
     putE cs
     putEStr rts
 
 putEventSpec (ProgramArgs cs as) = do
     let as' = unsep as
-    putE (fromIntegral (length as') + 4 :: Word16)
+    putE (fromIntegral (length as') + sz_capset :: Word16)
     putE cs
     mapM_ putE as'
 
 putEventSpec (ProgramEnv cs es) = do
     let es' = unsep es
-    putE (fromIntegral (length es') + 4 :: Word16)
+    putE (fromIntegral (length es') + sz_capset :: Word16)
     putE cs
     mapM_ putE es'
 
@@ -1196,6 +1288,10 @@ putEventSpec (UserMessage s) = do
     putE (fromIntegral (length s) :: Word16)
     mapM_ putE s
 
+putEventSpec (UserMarker s) = do
+    putE (fromIntegral (length s) :: Word16)
+    mapM_ putE s
+
 putEventSpec (UnknownEvent {}) = error "putEventSpec UnknownEvent"
 
 putEventSpec (InternString str id) = do
@@ -1241,6 +1337,20 @@ putEventSpec (MerReleaseThread thread_id) = do
 putEventSpec MerCapSleeping = return ()
 putEventSpec MerCallingMain = return ()
 
+putEventSpec PerfName{..} = do
+    putE (fromIntegral (length name) + sz_perf_num :: Word16)
+    putE perfNum
+    mapM_ putE name
+
+putEventSpec PerfCounter{..} = do
+    putE perfNum
+    putE tid
+    putE period
+
+putEventSpec PerfTracepoint{..} = do
+    putE perfNum
+    putE tid
+
 -- [] == []
 -- [x] == x\0
 -- [x, y, z] == x\0y\0
